/*
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Contacting the authors:
 *   Mario Arias:               mario.arias@gmail.com
 *   Javier D. Fernandez:       jfergar@infor.uva.es
 *   Miguel A. Martinez-Prieto: migumar2@infor.uva.es
 *
 */

#include <iomanip>
#include <algorithm>
#include <string>
#include <time.h>

#include <HDTVocabulary.hpp>
#include <RDFParser.hpp>

#include "ControlInformation.hpp"
#include "HDTFactory.hpp"
#include "BasicHDTExtended.hpp"
#include "../header/PlainHeader.hpp"
#include "../dictionary/PlainDictionary.hpp"
#include "../dictionary/KyotoDictionary.hpp"
#include "../dictionary/FourSectionDictionary.hpp"
#include "../dictionary/LiteralDictionary.hpp"

#include "../triples/TriplesList.hpp"
#include "../triples/TriplesKyoto.hpp"

#ifndef WIN32
#include "../triples/TripleListDisk.hpp"
#endif

#include "../triples/PlainTriples.hpp"
#include "../triples/CompactTriples.hpp"
#include "../triples/BitmapTriples.hpp"
#include "../triples/TripleOrderConvert.hpp"

#include "../third/gzstream.h"

#include "TripleIDStringIterator.hpp"

using namespace std;

namespace hdt {

BasicHDTExtended::BasicHDTExtended() :
		dict(NULL), triplesList(NULL) {
	createComponents();
}
BasicHDTExtended::BasicHDTExtended(HDTSpecification &spec) :
		dict(NULL), triplesList(NULL) {
	this->spec = spec;
	createComponents();
}



void BasicHDTExtended::startloadPartialDictionaries() {
	dict = getLoadDictionary();
	dict->startProcessing();
}

void BasicHDTExtended::endloadPartialDictionaries() {
	dict->stopProcessing(NULL);

	// Convert to final format
	dictionary->import(dict);

	delete dict;
}

void BasicHDTExtended::importDictionary(Dictionary* dict) {

	// Convert to final format
	dictionary->import(dict);

	delete dict;
}

void BasicHDTExtended::loadPartialDictionary(Dictionary * otherDict) {

	try {

		for (long long int i = 0; i < otherDict->getNsubjects(); i++) {
			string a = otherDict->idToString(i + 1, SUBJECT);
			dict->insert(a, SUBJECT);

		}

		for (long long int i = 0; i < otherDict->getNpredicates(); i++) {
			string a = otherDict->idToString(i + 1, PREDICATE);
			dict->insert(a, PREDICATE);
		}

		for (long long int i = 0; i < otherDict->getNobjects(); i++) {
			string a = otherDict->idToString(i + 1, OBJECT);
			dict->insert(a, OBJECT);

		}

	} catch (const char *e) {
		cout << "Catch exception dictionary: " << e << endl;
		delete dict;
		throw e;
	} catch (char *e) {
		cout << "Catch exception dictionary: " << e << endl;
		delete dict;
		throw e;
	}
}

void BasicHDTExtended::processMixTriple(string subject,int predicate,string object){
	TripleID ti;
	ti.setSubject(dictionary->stringToId(subject, SUBJECT));
	ti.setPredicate(predicate);
	ti.setObject(dictionary->stringToId(object, OBJECT));
		if (ti.isValid()) {
			triplesList->insert(ti);
		} else {
			cerr << "ERROR: Could not convert triple to IDS! " << endl << subject << " "<< predicate<<" "<<object<< endl << ti << endl;
		}
		//cout << "TripleID: " << ti << endl;
		char str[100];

		/*count++;
		if(pos>sizeBytes) {
			sizeBytes = pos;
		}*/
}

void BasicHDTExtended::processATriple(hdt::TripleString& triple) {
	TripleID ti;
	dictionary->tripleStringtoTripleID(triple, ti);
	if (ti.isValid()) {
		triplesList->insert(ti);
	} else {
		cerr << "ERROR: Could not convert triple to IDS! " << endl << triple << endl << ti << endl;
	}
	//cout << "TripleID: " << ti << endl;
	char str[100];

	/*count++;
	if(pos>sizeBytes) {
		sizeBytes = pos;
	}*/
}

void BasicHDTExtended::processIDTriple(hdt::TripleID ti) {
		triplesList->insert(ti);
		cout << "TripleID: " << ti << endl;
}


void BasicHDTExtended::stoploadPartialTriples() {
	try {
	triplesList->stopProcessing(NULL);

			// SORT & Duplicates
			TripleComponentOrder order = parseOrder(
					spec.get("triplesOrder").c_str());
			if (order == Unknown) {
				order = SPO;
			}
			triplesList->sort(order, NULL);

			triplesList->removeDuplicates(NULL);

			uint64_t totalOriginalSize=0; //fixme
			header->insert("_:statistics", HDTVocabulary::ORIGINAL_SIZE,
					totalOriginalSize);
	} catch (const char *e) {
			cout << "Catch exception triples" << e << endl;
			delete triplesList;
			throw e;
		} catch (char *e) {
			cout << "Catch exception triples" << e << endl;
			delete triplesList;
			throw e;
		}
		if (triples->getType() == triplesList->getType()) {
			delete triples;
			triples = triplesList;
		} else {
			try {
				triples->load(*triplesList, NULL);
			} catch (const char* e) {
				delete triplesList;
				throw e;
			}
			delete triplesList;
		}
}
void BasicHDTExtended::startloadPartialTriples() {
	triplesList = new TriplesList(spec);
    triplesList->startProcessing(NULL);
}


void BasicHDTExtended::endloadFromPartial(string baseUri) {
	try {
		// Make sure that URI starts and ends with <>
		if (baseUri.at(0) != '<')
			baseUri = '<' + baseUri;
		if (baseUri.at(baseUri.length() - 1) != '>')
			baseUri.append(">");

	fillHeader(baseUri);
	} catch (const char *e) {
			cout << "Catch exception load: " << e << endl;
			deleteComponents();
			createComponents();
			throw e;
		} catch (char *e) {
			cout << "Catch exception load: " << e << endl;
			deleteComponents();
			createComponents();
			throw e;
		}
}



}
