/*
 * File: HDTEnums.cpp
 * Last modified: $Date: 2012-08-13 23:00:07 +0100 (lun, 13 ago 2012) $
 * Revision: $Revision: 222 $
 * Last modified by: $Author: mario.arias $
 *
 * Copyright (C) 2012, Mario Arias, Javier D. Fernandez, Miguel A. Martinez-Prieto
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 * Contacting the authors:
 *   Mario Arias:               mario.arias@gmail.com
 *   Javier D. Fernandez:       jfergar@infor.uva.es
 *   Miguel A. Martinez-Prieto: migumar2@infor.uva.es
 *
 */

#include <HDT.hpp>
#include <HDTManager.hpp>
#include <signal.h>

#include <getopt.h>
#include <string.h>
#include <string>
#include <iostream>
#include <fstream>
#include "../src/util/StopWatch.hpp"

using namespace hdt;
using namespace std;

int interruptSignal = 0;

void signalHandler(int sig) {
	interruptSignal = 1;
}

void help() {
	cout << "$ hdtSearch [options] <hdtfile> " << endl;
	cout << "\t-h\t\t\tThis help" << endl;
	cout << "\t-q\t<query>\t\tLaunch query and exit." << endl;
	cout << "\t-Q\t<queryFile>\t\tLaunch queries and exit." << endl;
	cout << "\t-n\t\t\tUse no indexes (No HDT-Foq)." << endl;
	cout << "\t-o\t<output>\tSave query output to file." << endl;
	cout << "\t-T\t<QueryTimeFile>\tSave query time to file." << endl;
	cout << "\t-m\t\t\tDo not show results, just measure query time." << endl;

	//cout << "\t-v\tVerbose output" << endl;
}

void iterate(HDT *hdt, char *query, ostream &out, bool measure) {
	TripleString tripleString;
	tripleString.read(query);

	const char *subj = tripleString.getSubject().c_str();
	const char *pred = tripleString.getPredicate().c_str();
	const char *obj = tripleString.getObject().c_str();
	if (strcmp(subj, "?") == 0) {
		subj = "";
	}
	if (strcmp(pred, "?") == 0) {
		pred = "";
	}
	if (strcmp(obj, "?") == 0) {
		obj = "";
	}

#if 0
	cout << "Subject: |" << subj << "|" << endl;
	cout << "Predicate: |" << pred << "|" << endl;
	cout << "Object: |" << obj << "|" << endl;
#endif

	try {
		IteratorTripleString *it = hdt->search(subj, pred, obj);

		StopWatch st;
		unsigned int numTriples = 0;
		while (it->hasNext() && interruptSignal == 0) {
			TripleString *ts = it->next();
			if (!measure)
				out << *ts << endl;
			numTriples++;
		}
		cout << numTriples << " results in " << st << endl;
		delete it;

		interruptSignal = 0;	// Interrupt caught, enable again.
	} catch (char *e) {
		cerr << e << endl;
	}

}

class ConvertProgress: public ProgressListener {
private:
public:
	virtual ~ConvertProgress() {
	}

	void notifyProgress(float level, const char *section) {
		cout << section << ": " << level << " %";
		cout << "\r " << section << ": " << level
				<< " %                      \r";
		cout.flush();
	}

};

int main(int argc, char **argv) {
	int c;
	string query, inputFile, outputFile, queryFile, queryTimeFile,
			loadingTimeFile;
	bool measure = false;
	bool indexes = true;

	while ((c = getopt(argc, argv, "hq:o:mQ:nT:")) != -1) {
		switch (c) {
		case 'h':
			help();
			break;
		case 'q':
			query = optarg;
			break;
		case 'Q':
			queryFile = optarg;
			break;
		case 'o':
			outputFile = optarg;
			break;
		case 'T':
			queryTimeFile = optarg;
			break;
		case 'm':
			measure = true;
			break;
		case 'n':
			indexes = false;
			break;
		default:
			cout << "ERROR: Unknown option" << endl;
			help();
			return 1;
		}
	}

	if (argc - optind < 1) {
		cout << "ERROR: You must supply an HDT File" << endl << endl;
		help();
		return 1;
	}

	inputFile = argv[optind];

	try {
		ConvertProgress prog;
		HDT *hdt;

		StopWatch globalTimer;
		StopWatch loadTimer;
		if (indexes)
			hdt = HDTManager::mapIndexedHDT(inputFile.c_str(), &prog);
		else
			hdt = HDTManager::mapHDT(inputFile.c_str(), &prog);

		loadTimer.stop();
		ostream *out;
		ofstream outF;

		if (outputFile != "") {
			outF.open(outputFile.c_str());
			out = &outF;
		} else {
			out = &cout;
		}

		ostream *outQueryTime;
		ofstream outFQueryTime;

		if (queryTimeFile != "") {
			outFQueryTime.open(queryTimeFile.c_str());
			outQueryTime = &outFQueryTime;
		} else {
			outQueryTime = &cout;
		}
		StopWatch queryTimer;


		if (query != "") {
			// Supplied query, search and exit.
			iterate(hdt, (char*) query.c_str(), *out, measure);
		} else if (queryFile != "") {
			//read queries
			std::ifstream inqueryfile(queryFile.c_str());
			std::string queryLine;

			int numQueries=0;
			while (std::getline(inqueryfile, queryLine)) {
				iterate(hdt, (char*) queryLine.c_str(), *out, measure);
				numQueries++;
			}


			cout << "Load time: " << loadTimer << endl;
			cout<< "Query time: " << queryTimer << endl;
			cout<< "Total processing time: " << globalTimer << endl;
			if (queryTimeFile != "") {
				queryTimer.stop();
				globalTimer.stop();
				*outQueryTime << "Load time seconds: " << ((double)loadTimer.getReal()/1000000) << endl;
				*outQueryTime << "Query time: " << ((double)queryTimer.getReal()/1000000) << endl;
				*outQueryTime << "Mean Q time: " << ((double)queryTimer.getReal()/1000000/numQueries) << endl;
				*outQueryTime << "Total processing time: " << ((double)globalTimer.getReal()/1000000) << endl;
			}


		//	cout << "Clock(" << globalTimer.getRealStr();
		//	cout << ")  User(" << globalTimer.getUserStr();
		//	cout << ")  System(" << globalTimer.getSystemStr() << ")" << endl;
		} else {
			// No supplied query, show terminal.
			char line[1024 * 10];

			signal(SIGINT, &signalHandler);
			cout << "                                                 \r>> ";
			while (cin.getline(line, 1024 * 10)) {
				if (strcmp(line, "exit") == 0 || strcmp(line, "quit") == 0) {
					break;
				}
				if (strlen(line) == 0 || strcmp(line, "help") == 0) {
					cout
							<< "Please type Triple Search Pattern, using '?' for wildcards. e.g "
							<< endl;
					cout << "   http://www.somewhere.com/mysubject ? ?" << endl;
					cout
							<< "Interrupt with Control+C. Type 'exit', 'quit' or Control+D to exit the shell."
							<< endl;
					cout << ">> ";
					continue;
				}

				iterate(hdt, line, *out, measure);

				cout << ">> ";
			}
		}

		if (outputFile != "") {
			outF.close();
		}

		delete hdt;
	} catch (char *e) {
		cout << "ERROR: " << e << endl;
	} catch (const char *e) {
		cout << "ERROR: " << e << endl;
	}
}

